/*
 * Copyright (C) 2010 hidetake.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.hidetake.jsonizer;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.hidetake.jsonizer.JsonWriter;
import org.hidetake.jsonizer.PropertyGetter;
import org.junit.Before;
import org.junit.Test;

import static org.junit.Assert.*;

public class JsonWriterTest
{

	public static class Hoge
	{
		private int count;
		private String message;
		public int getCount()
		{
			return count;
		}
		public void setCount(int count)
		{
			this.count = count;
		}
		public String getMessage()
		{
			return message;
		}
		public void setMessage(String message)
		{
			this.message = message;
		}
	}
	
	public static class Fuga
	{
		private final String name;
		private final List<Hoge> hoges = new ArrayList<Hoge>();
		public Fuga(String name)
		{
			this.name = name;
		}
		public String getName()
		{
			return name;
		}
		public List<Hoge> getHoges()
		{
			return hoges;
		}
		public boolean add(Hoge e)
		{
			return hoges.add(e);
		}
		public void clear()
		{
			hoges.clear();
		}
	}
	
	public static class Piyo
	{
		private int id;
		private Piyo child;
		public int getId()
		{
			return id;
		}
		public void setId(int id)
		{
			this.id = id;
		}
		public Piyo getChild()
		{
			return child;
		}
		public void setChild(Piyo child)
		{
			this.child = child;
		}
	}
	
	public static class HogeGetter extends PropertyGetter<Hoge>
	{
		@Override
		public void get(Hoge m)
		{
			property(m.getCount()).as("count");
			property("msg", m.getMessage());
		}
	}
	
	public static class HogePartialGetter extends PropertyGetter<Hoge>
	{
		@Override
		public void get(Hoge m)
		{
			property("msg", m.getMessage());
		}
	}

	private StringBuilder result;
	
	@Before
	public void setUp()
	{
		result = new StringBuilder();
	}

	@Test
	public void _リテラル_int() throws Exception
	{
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(new Object[]{1});
		assertEquals("[1]", result.toString());
	}

	@Test
	public void _リテラル_long() throws Exception
	{
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(new Object[]{2L});
		assertEquals("[2]", result.toString());
	}

	@Test
	public void _リテラル_boolean() throws Exception
	{
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(new Object[]{false});
		assertEquals("[false]", result.toString());
	}

	@Test
	public void _リテラル_char() throws Exception
	{
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(new Object[]{'x'});
		assertEquals("[\"x\"]", result.toString());
	}

	@Test
	public void _リテラル_float() throws Exception
	{
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(new Object[]{3.0f});
		assertEquals("[3.0]", result.toString());
	}

	@Test
	public void _リテラル_double() throws Exception
	{
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(new Object[]{4.0});
		assertEquals("[4.0]", result.toString());
	}

	@Test
	public void _リテラル_NaN() throws Exception
	{
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(new Object[]{Double.NaN});
		assertEquals("[NaN]", result.toString());
	}

	@Test
	public void _リテラル_Inf() throws Exception
	{
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(new Object[]{Double.POSITIVE_INFINITY});
		assertEquals("[Infinity]", result.toString());
	}

	@Test
	public void _リテラル_MInf() throws Exception
	{
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(new Object[]{Double.NEGATIVE_INFINITY});
		assertEquals("[-Infinity]", result.toString());
	}

	@Test
	public void _リテラル_byte() throws Exception
	{
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(new Object[]{(byte)32});
		assertEquals("[32]", result.toString());
	}

	@Test
	public void _リテラル_short() throws Exception
	{
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(new Object[]{(short)64});
		assertEquals("[64]", result.toString());
	}

	@Test
	public void _リテラル_String() throws Exception
	{
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(new Object[]{"Hoge"});
		assertEquals("[\"Hoge\"]", result.toString());
	}

	@Test
	public void _リテラル_String_empty() throws Exception
	{
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(new Object[]{"Hoge"});
		assertEquals("[\"Hoge\"]", result.toString());
	}

	@Test
	public void _リテラル_String_改行() throws Exception
	{
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(new Object[]{""});
		assertEquals("[\"\"]", result.toString());
	}

	@Test
	public void _リテラル_String_改行_first() throws Exception
	{
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(new Object[]{"\nxxxxxx"});
		assertEquals("[\"\\nxxxxxx\"]", result.toString());
	}

	@Test
	public void _リテラル_String_改行_mid() throws Exception
	{
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(new Object[]{"xxx\nxxx"});
		assertEquals("[\"xxx\\nxxx\"]", result.toString());
	}

	@Test
	public void _リテラル_String_改行_last() throws Exception
	{
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(new Object[]{"xxxxxx\n"});
		assertEquals("[\"xxxxxx\\n\"]", result.toString());
	}

	@Test
	public void _リテラル_char_改行() throws Exception
	{
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(new Object[]{'\n'});
		assertEquals("[\"\\n\"]", result.toString());
	}

	@Test
	public void _リテラル_String_バックスラッシュ() throws Exception
	{
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(new Object[]{"\\"});
		assertEquals("[\"\\\\\"]", result.toString());
	}

	@Test
	public void _リテラル_char_バックスラッシュ() throws Exception
	{
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(new Object[]{'\\'});
		assertEquals("[\"\\\\\"]", result.toString());
	}

	@Test
	public void _リテラル_String_ダブルクォート() throws Exception
	{
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(new Object[]{"\""});
		assertEquals("[\"\\\"\"]", result.toString());
	}

	@Test
	public void _リテラル_char_ダブルクォート() throws Exception
	{
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(new Object[]{'\"'});
		assertEquals("[\"\\\"\"]", result.toString());
	}

	@Test
	public void _配列() throws Exception
	{
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(new Integer[]{1,2,3});
		assertEquals("[1,2,3]", result.toString());
	}

	@Test
	public void _配列2() throws Exception
	{
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(new Object[]{1,2L,"Hoge",false,'x',3.0f,4.0,(byte)32,(short)64});
		assertEquals("[1,2,\"Hoge\",false,\"x\",3.0,4.0,32,64]", result.toString());
	}

	@Test
	public void _配列_empty() throws Exception
	{
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(new Object[]{});
		assertEquals("[]", result.toString());
	}

	@Test
	public void _配列_null() throws Exception
	{
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(new Object[]{null});
		assertEquals("[null]", result.toString());
	}

	@Test
	public void _配列_null2() throws Exception
	{
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(new Integer[]{null,10});
		assertEquals("[null,10]", result.toString());
	}

	@Test
	public void _配列_ネスト() throws Exception
	{
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(new Object[]{1,new Object[]{100,"Hoge"},2});
		assertEquals("[1,[100,\"Hoge\"],2]", result.toString());
	}

	@Test
	public void _配列_Object() throws Exception
	{
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(new Object[]{this});
		String expected = "[\""+this.getClass().getName();
		assertEquals(expected, result.toString().substring(0, expected.length()));
	}

	@Test
	public void _リスト() throws Exception
	{
		List<Integer> list = new ArrayList<Integer>();
		list.add(1);
		list.add(2);
		list.add(3);
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(list);
		assertEquals("[1,2,3]", result.toString());
	}

	@Test
	public void _リスト2() throws Exception
	{
		List<Object> list = new ArrayList<Object>();
		list.add(1);
		list.add(2);
		list.add("Hoge");
		list.add(false);
		list.add('x');
		list.add(3.0f);
		list.add(4.0);
		list.add((byte)32);
		list.add((short)64);
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(list);
		assertEquals("[1,2,\"Hoge\",false,\"x\",3.0,4.0,32,64]", result.toString());
	}

	@Test
	public void _リスト_empty() throws Exception
	{
		List<Object> list = new ArrayList<Object>();
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(list);
		assertEquals("[]", result.toString());
	}

	@Test
	public void _リスト_null() throws Exception
	{
		List<Object> list = new ArrayList<Object>();
		list.add(null);
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(list);
		assertEquals("[null]", result.toString());
	}

	@Test
	public void _リスト_null2() throws Exception
	{
		List<Object> list = new ArrayList<Object>();
		list.add(null);
		list.add(10);
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(list);
		assertEquals("[null,10]", result.toString());
	}

	@Test
	public void _リスト_ネスト() throws Exception
	{
		List<Object> list = new ArrayList<Object>();
		list.add(1);
		List<Object> list2 = new ArrayList<Object>();
		list2.add(100);
		list2.add("Hoge");
		list.add(list2);
		list.add(2);
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(list);
		assertEquals("[1,[100,\"Hoge\"],2]", result.toString());
	}

	@Test
	public void _リスト_Object() throws Exception
	{
		List<Object> list = new ArrayList<Object>();
		list.add(this);
		JsonWriter jsonWriter = new JsonWriter(result);
		jsonWriter.write(list);
		String expected = "[\""+this.getClass().getName();
		assertEquals(expected, result.toString().substring(0, expected.length()));
	}
	
	@Test
	public void _ハッシュ() throws Exception
	{
		Hoge hoge = new Hoge();
		hoge.setCount(100);
		hoge.setMessage("OK");
		
		Map<Class<?>, PropertyGetter<?>> map = new HashMap<Class<?>, PropertyGetter<?>>();
		map.put(Hoge.class, new HogeGetter());
		JsonWriter jsonWriter = new JsonWriter(result, map);
		jsonWriter.write(hoge);
		assertEquals("{\"count\":100,\"msg\":\"OK\"}", result.toString());
	}
	
	@Test
	public void _ハッシュ_null() throws Exception
	{
		Hoge hoge = new Hoge();
		hoge.setCount(100);
		hoge.setMessage(null);
		
		Map<Class<?>, PropertyGetter<?>> map = new HashMap<Class<?>, PropertyGetter<?>>();
		map.put(Hoge.class, new HogeGetter());
		JsonWriter jsonWriter = new JsonWriter(result, map);
		jsonWriter.write(hoge);
		assertEquals("{\"count\":100,\"msg\":null}", result.toString());
	}
	
	@Test
	public void _ハッシュ_partial() throws Exception
	{
		Hoge hoge = new Hoge();
		hoge.setCount(100);
		hoge.setMessage(null);
		
		Map<Class<?>, PropertyGetter<?>> map = new HashMap<Class<?>, PropertyGetter<?>>();
		map.put(Hoge.class, new HogePartialGetter());
		JsonWriter jsonWriter = new JsonWriter(result, map);
		jsonWriter.write(hoge);
		assertEquals("{\"msg\":null}", result.toString());
	}
	
	@Test
	public void _ハッシュ_empty() throws Exception
	{
		Hoge hoge = new Hoge();
		hoge.setCount(100);
		hoge.setMessage(null);
		
		Map<Class<?>, PropertyGetter<?>> map = new HashMap<Class<?>, PropertyGetter<?>>();
		map.put(Hoge.class, PropertyGetter.emptyGetter());
		JsonWriter jsonWriter = new JsonWriter(result, map);
		jsonWriter.write(hoge);
		assertEquals("{}", result.toString());
	}
	
	@Test
	public void _リスト_object() throws Exception
	{
		List<Object> list = new ArrayList<Object>();
		Hoge hoge = new Hoge();
		hoge.setCount(100);
		hoge.setMessage("OK");
		list.add(hoge);
		list.add(0);
		
		Map<Class<?>, PropertyGetter<?>> map = new HashMap<Class<?>, PropertyGetter<?>>();
		map.put(Hoge.class, new HogeGetter());
		JsonWriter jsonWriter = new JsonWriter(result, map);
		jsonWriter.write(list);
		assertEquals("[{\"count\":100,\"msg\":\"OK\"},0]", result.toString());
	}
	
	@Test
	public void _リスト_ハッシュ() throws Exception
	{
		Fuga fuga = new Fuga("some");
		{
			Hoge hoge = new Hoge();
			hoge.setCount(100);
			hoge.setMessage(null);
			fuga.add(hoge);
		}
		{
			Hoge hoge = new Hoge();
			hoge.setCount(50);
			hoge.setMessage("OK");
			fuga.add(hoge);
		}
		
		Map<Class<?>, PropertyGetter<?>> map = new HashMap<Class<?>, PropertyGetter<?>>();
		map.put(Hoge.class, new HogeGetter());
		map.put(Fuga.class, new PropertyGetter<Fuga>(){
			@Override
			public void get(Fuga m)
			{
				property("name", m.getName());
				property("list", m.getHoges());
			}
		});
		
		JsonWriter jsonWriter = new JsonWriter(result, map);
		jsonWriter.write(fuga);
		assertEquals("{\"name\":\"some\",\"list\":[" +
				"{\"count\":100,\"msg\":null}," +
				"{\"count\":50,\"msg\":\"OK\"}]}",
				result.toString());
	}
	
	@Test
	public void _ハッシュ_ネスト() throws Exception
	{
		Piyo piyo1 = new Piyo();
		piyo1.setId(1);
		{
			Piyo piyo2 = new Piyo();
			piyo1.setChild(piyo2);
			piyo2.setId(2);
			{
				Piyo piyo3 = new Piyo();
				piyo2.setChild(piyo3);
				piyo3.setId(3);
				piyo3.setChild(null);
			}
		}
		
		Map<Class<?>, PropertyGetter<?>> map = new HashMap<Class<?>, PropertyGetter<?>>();
		map.put(Piyo.class, new PropertyGetter<Piyo>()
		{
			@Override
			public void get(Piyo m)
			{
				property("id", m.getId());
				property("child", m.getChild());
			}
		});
		
		JsonWriter jsonWriter = new JsonWriter(result, map);
		jsonWriter.write(piyo1);
		assertEquals("{\"id\":1,\"child\":{\"id\":2,\"child\":{\"id\":3,\"child\":null}}}", result.toString());
	}

	@Test(expected = IllegalStateException.class)
	public void _PropertyGetter不正() throws Exception
	{
		Piyo piyo = new Piyo();
		piyo.setId(1);
		piyo.setChild(null);
		Map<Class<?>, PropertyGetter<?>> map = new HashMap<Class<?>, PropertyGetter<?>>();
		map.put(Piyo.class, new PropertyGetter<Piyo>()
		{
			@Override
			public void get(Piyo m)
			{
				property(m.getId()); //forgot as()
				property(m.getChild()).as("child");
			}
		});
		
		JsonWriter jsonWriter = new JsonWriter(result, map);
		jsonWriter.write(piyo);
	}

}
